#!/usr/bin/python

"""
Regrids ghost's fortran native out
Author :  Alexandre CAMERON
Email  :  alexandre.cameron@ens.fr
Input  :  radical e.g.: vx.0001.out
Command:  python regrid.py
Default:  double resolution 
          ./outs/vxyz.0001.out -> ./outs/vxyz.9999.out
Caution:  Make sure you have the correct path
          for python on the first line after #!
          check shell command: >  which python
"""

import numpy as np
import sys
import middle as md

#######
### DFT
#######

module="numpy"#"numpy"#
### DFT numpy
if module=="numpy" :
    import numpy.fft as dft
### FFTW 
elif module=="pyfftw":
    import pyfftw.interfaces.numpy_fft as dft

def HermiteProd ( vec1, vec2 ) : 
    hn = (np.real(vec1*((vec2).conjugate()))).sum()
    return hn
###
def ParPlan ( vec, four ) :
    num = HermiteProd ( vec, vec )
    den = HermiteProd ( four, four )
    pp = np.sqrt( num  / den )
    return pp
###
def NormFT ( Nres ) :
    One  = np.ones( (Nres, Nres, Nres ) )
    Fone = dft.fftn(One)
    norm = ParPlan(One, Fone)
    return norm
###
def DirectFT ( Vx ) :
    norm = NormFT(len(Vx))
    FVx  = norm*dft.fftn(Vx)
    return ( FVx )
###
def NmIvFT ( Nres ) :
    One  = np.ones( (Nres, Nres, Nres ) )
    Fone = dft.ifftn(One)
    norm = ParPlan(One, Fone)
    return norm
###
def InvertFT ( Vx ) :
    norm = NmIvFT(len(Vx))
    FVx  = norm*dft.ifftn(Vx)
    return ( np.real(FVx) )
###
def fftn ( Vx ) :
    return dft.fftn(Vx)
###
def ifftn ( Vx ) :
    return dft.ifftn(Vx)

##############
### IO Fortran
##############

def FiNaFor ( ax='x', numf=1, way='./outs/', 
              var='v', ext='.out'):
    """
    File Name Fortran : name of the file to read
    """
    name = ( str(way) + str(var) + str(ax) + '.' + 
             '%0.4i' % (int(numf)) + str(ext) )
    return name
###
def Extract ( ftype, name ) :
    file = open(name, 'r')
    # Unknown endianness
    vic  = np.fromfile(file,dtype=str(ftype),count=-1) 
    nres = (np.log(len(vic))/np.log(2))/3
    if ( nres != np.ceil(nres) ) :
        errMess = 'readFor.py: res not iso -> check type' 
        sys.exit(errMess)
    Nres = int(np.power(2,nres))
    NresX= Nres ; NresY= Nres ; NresZ= Nres
    Vec  = np.reshape(vic, (NresX, NresY, NresZ))
    return Vec
###
def Xtrak3D ( ftype='float64', ninf=1, 
              way='./outs/',  var='v', ext='.out', verbose=False) :
    Vx  = Extract( ftype, FiNaFor('x', ninf, way, var, ext) )
    if verbose : print 'Extract X done'
    Vy  = Extract( ftype, FiNaFor('y', ninf, way, var, ext) )
    if verbose : print 'Extract Y done'
    Vz  = Extract( ftype, FiNaFor('z', ninf, way, var, ext) )
    if verbose : print 'Extract Z done'
    return ( Vx, Vy, Vz )
###
def Dump ( nV, ftype, name ) :
    file = open(name, 'w' )
    ((np.transpose(nV)).astype(ftype)).tofile(file)
    # The transposition is there to go 
    # from C ordering to Fortran ordering 
    # Other option : file.write((nV.astype(ftype)).tobytes(order='F'))
    file.close()
    return 
###
def DumpAll ( nVx, nVy, nVz, ftype, verbose=False, 
              notf=9999, way='./outs/',  var='v', ext='.out' ) :
    Dump ( nVx, ftype, FiNaFor('x', notf, way, var, ext) )
    if verbose : print 'Dump X done'
    Dump ( nVy, ftype, FiNaFor('y', notf, way, var, ext) )
    if verbose : print 'Dump Y done'
    Dump ( nVz, ftype, FiNaFor('z', notf, way, var, ext) )
    if verbose : print 'Dump Z done'
    return

##########
### Resize
##########

def Resize ( F, OldRes, NewRes, ftype ) :
    nF = np.zeros((NewRes,NewRes,NewRes ), dtype=np.dtype(F[0,0,0]))
    MinRes = np.min([OldRes, NewRes])
    res = MinRes/2
    for iter in xrange(res) :
        for jter in xrange(res) :
            for kter in xrange(res) :
                nF[iter,jter,kter] = F[iter,jter,kter] 
                #
                nF[-1-iter,jter,kter] = F[-1-iter,jter,kter] 
                nF[iter,-1-jter,kter] = F[iter,-1-jter,kter] 
                nF[iter,jter,-1-kter] = F[iter,jter,-1-kter] 
                #
                nF[-1-iter,-1-jter,kter] = F[-1-iter,-1-jter,kter] 
                nF[iter,-1-jter,-1-kter] = F[iter,-1-jter,-1-kter] 
                nF[-1-iter,jter,-1-kter] = F[-1-iter,jter,-1-kter] 
                #
                nF[-1-iter,-1-jter,-1-kter] = F[-1-iter,-1-jter,-1-kter] 
    return nF
###
def ResizeAll ( FVx, FVy, FVz, OldRes, NewRes, 
                ftype='float64', verbose=False ):
    nFVx = Resize( FVx, OldRes, NewRes, ftype )
    if verbose : print 'Resize X done'
    nFVy = Resize( FVy, OldRes, NewRes, ftype )
    if verbose : print 'Resize Y done'
    nFVz = Resize( FVz, OldRes, NewRes, ftype )
    if verbose : print 'Resize Z done'
    return ( nFVx, nFVy, nFVz ) 

##########
### Middle
##########

def Middle ( V, OldRes, NewRes, ftype ) :
    nV = np.zeros((NewRes,NewRes,NewRes ), dtype=np.dtype(V[0,0,0]))
    MinRes = np.min([OldRes, NewRes])
    res = MinRes
    for iter in xrange(res) :
        for jter in xrange(res) :
            for kter in xrange(res) :
                nV[2*iter,2*jter,2*kter] = V[iter,jter,kter] 
                #
                nV[2*iter+1,2*jter,2*kter] = 0.5*(V[iter,jter,kter] + V[(iter+1)%res,jter,kter] )
                nV[2*iter,2*jter+1,2*kter] = 0.5*(V[iter,jter,kter] + V[iter,(jter+1)%res,kter] )
                nV[2*iter,2*jter,2*kter+1] = 0.5*(V[iter,jter,kter] + V[iter,jter,(kter+1)%res] )
                #
                nV[2*iter+1,2*jter+1,2*kter] = 0.5*(V[iter,jter,kter] + V[(iter+1)%res,(jter+1)%res,kter] )
                nV[2*iter,2*jter+1,2*kter+1] = 0.5*(V[iter,jter,kter] + V[iter,(jter+1)%res,(kter+1)%res] )
                nV[2*iter+1,2*jter,2*kter+1] = 0.5*(V[iter,jter,kter] + V[(iter+1)%res,jter,(kter+1)%res] )
                #
                nV[2*iter+1,2*jter+1,2*kter+1]= 0.5*(V[iter,jter,kter] + V[(iter+1)%res,(jter+1)%res,(kter+1)%res])
                #
        if iter%10 ==0 :
            sys.stdout.write("\r"+ str(iter) + '/' + str(res) )
            sys.stdout.flush()
    return nV
###
def MiddleAll ( Vx, Vy, Vz, OldRes, NewRes, ftype, verbose=False ) :
    nVx = md.Middle ( Vx, int(OldRes), int(NewRes), str(ftype) )
    if verbose : print "Middle X done"
    nVy = md.Middle ( Vy, int(OldRes), int(NewRes), str(ftype) )
    if verbose : print "Middle Y done"
    nVz = md.Middle ( Vz, int(OldRes), int(NewRes), str(ftype) )
    if verbose : print "Middle Z done"
    return (nVx, nVy, nVz)
    
#######
### DFT
#######

def MakeFT ( Vx, Vy, Vz, OldRes, NewRes, ftype, verbose=False ) :
    norm= NormFT(OldRes) / NormFT(NewRes) 
    FVx = norm*fftn(Vx)
    if verbose : print 'FT X done'
    FVy = norm*fftn(Vy)
    if verbose : print 'FT Y done'
    FVz = norm*fftn(Vz)
    if verbose : print 'FT Z done'
    return ( FVx, FVy, FVz )
###
def MkIvFT ( nFVx, nFVy, nFVz, OldRes, NewRes, ftype, verbose=False ) :
    norm= NmIvFT(NewRes) / NmIvFT(OldRes)
    nVx = np.real( norm*ifftn(nFVx) )
    if verbose : print 'IvFT X done'
    nVy = np.real( norm*ifftn(nFVy) )
    if verbose : print 'IvFT Y done'
    nVz = np.real( norm*ifftn(nFVz) )
    if verbose : print 'IvFT Z done'
    return ( nVx, nVy, nVz )

###########
### AuxMain
###########

def OneFlag ( flag, args, lbl ): 
    if ( len(args) > lbl) :
        flag = str(args[lbl])
        return ( flag, lbl+1)
    else:
        return ( flag, lbl )
###
def Arg2Flag ( args ) :
    NewRes = None 
    ftype='float64'; ninf=1; notf=9999;  
    way='./outs/'; var='v'; ext='.out' ; 
    lbl=0
    #
    (ftype,lbl) = OneFlag ( ftype, args, lbl )
    (nr,lbl) = OneFlag ( NewRes, args, lbl )
    if nr==None : NewRes=None
    else : NewRes = int(nr)        
    (ninf, lbl) = OneFlag ( ninf, args, lbl )
    (notf, lbl) = OneFlag ( notf, args, lbl )
    (way, lbl) = OneFlag ( way, args, lbl )
    (var, lbl) = OneFlag ( var, args, lbl )
    (ext, lbl) = OneFlag ( ext, args, lbl )
    ninf = str(ninf)
    return ( ftype, NewRes, ninf, notf, way, var, ext)
###
def ArgsMain ( argsys, ftype ) :
   args = []
   args.append(ftype)
   if (len(argsys) > 1 ):
       for index in range(1, len(argsys) ) :
           args.append( argsys[index] )
   return args

########
### MAIN
########

def DoIt ( args, FFT=False, verbose=False ) :
    """
    args: [ftype, NewRes, ninf, way, var, ext]
    """
    ( ftype, NewRes, ninf, notf, way, var, ext ) = Arg2Flag (args)
    fin = FiNaFor('xyz', ninf, way, var, ext)
    fot = FiNaFor('xyz', notf, way, var, ext)
    print 'Regriding ' + fin + ' -> ' + fot
    if verbose : print ( ftype, NewRes, ninf, way, var, ext )
    if verbose : print 'Arg2Flag done'
    ( Vx, Vy, Vz ) = Xtrak3D ( ftype, ninf, way, var, ext, verbose ) 
    if verbose : print 'Xtrak3D done'
    OldRes = len(Vx)
    if NewRes==None : NewRes = 2*OldRes
    print 'Converting ' + str(OldRes) + ' -> ' + str(NewRes)
    if FFT :
        print "FFT method"
        ( FVx, FVy, FVz ) = MakeFT ( Vx, Vy, Vz, OldRes, NewRes, ftype, verbose )    
        if verbose : print 'MakeFT done'
        ( nFVx, nFVy, nFVz ) = ResizeAll ( FVx, FVy, FVz, OldRes, NewRes, ftype, verbose )
        if verbose : print 'ResizeAll done'
        ( nVx, nVy, nVz ) = MkIvFT ( nFVx, nFVy, nFVz, OldRes, NewRes, ftype, verbose )
        if verbose : print 'MkIvFT done'
    else :
        print "Middle method"
        ( nVx, nVy, nVz ) = MiddleAll ( Vx, Vy, Vz, OldRes, NewRes, ftype, verbose )
        if verbose : print 'MiddleAll done'
    DumpAll( nVx, nVy, nVz, ftype, verbose, notf, way, var, ext )
    if verbose : print 'DumpAll done'
    print 'Conversion done'    
    return

if __name__ == '__main__':
    ftype = 'float64'
    verbose = True # False
    FFT = False
    args = ArgsMain ( sys.argv, ftype )
    DoIt( args, FFT, verbose )
